@@ -1458,10 +1458,19 @@ _mesa_GetTextureSubImage(GLuint texture, GLint level,
 {
    GET_CURRENT_CONTEXT(ctx);
    static const char *caller = "glGetTextureSubImage";
-   struct gl_texture_object *texObj =
-      _mesa_lookup_texture_err(ctx, texture, caller);
+   struct gl_texture_object *texObj = NULL;
+
+   if (texture > 0)
+      texObj = _mesa_lookup_texture(ctx, texture);
 
    if (!texObj) {
+      _mesa_error(ctx, GL_INVALID_VALUE, "%s(texture)", caller);
+      return;
+   }
+
+   if (!legal_getteximage_target(ctx, texObj->Target, true)) {
+      _mesa_error(ctx, GL_INVALID_OPERATION,
+                  "%s(buffer/multisample texture)", caller);
       return;
    }
 
@@ -1767,10 +1776,13 @@ _mesa_GetCompressedTextureSubImage(GLuint texture, GLint level,
 {
    GET_CURRENT_CONTEXT(ctx);
    static const char *caller = "glGetCompressedTextureImage";
-   struct gl_texture_object *texObj;
+   struct gl_texture_object *texObj = NULL;
+
+   if (texture > 0)
+      texObj = _mesa_lookup_texture(ctx, texture);
 
-   texObj = _mesa_lookup_texture_err(ctx, texture, caller);
    if (!texObj) {
+      _mesa_error(ctx, GL_INVALID_VALUE, "%s(texture)", caller);
       return;
    }
 
